/* SymLinkDialog.cpp
Defines the main dialog plus most of the functionality
(C) 2002 Niall Douglas under the GNU General Public Licence
*/

#include "symlinkdialog.h"
#include "help.h"
#include <qdragobject.h>
#include <qpushbutton.h>
#include <qlineedit.h>
#include <qlabel.h>
#include <qfiledialog.h>
#include <qmessagebox.h>
#include <qtextbrowser.h>
#include <qdir.h>
#include "FSLinks.h"

int Support_StringToCharArray(QChar *dest, const QString &src, int max)
{
	int len=src.length();
	if(len>max-1)
	{
		len=max-1;
		qWarning("Support_StringToCharArray - string truncated");
	}
	memcpy(dest, src.unicode(), len*sizeof(QChar));
	dest[len]=0;
	return len;
}

QString &Support_CharArrayToString(QString &dest, const QChar *src)
{
	return dest.replace(0, dest.length(), src, wcslen((wchar_t *) src));
}

SymLinkDialog::SymLinkDialog( QWidget* parent, const char* name, bool modal, WFlags f )
	: SymLinkDialogBase( parent, name, modal, f )
{
	help=new Help(this, 0, FALSE);
	DestPath->installEventFilter(this);
	DeleteDrop->installEventFilter(this);
	help->Display->setText("<html><body>"
		"<h1>SymLink help</h1>v1.02 (13th July 2002)<br><i>(C) 2002 Niall Douglas</i><p>"
		"NTFS 5 on Microsoft Windows finally has partial support for symbolic links with which you can finally map arbitrary "
		"items from any location on the local machine to appear as though they were in a directory on an NTFS 5 volume. NTFS 5 is present "
		"with Windows 2000 or later<p>"
		"The most common and useful ability of this is to unify multiple partitions to appear on a single drive. This "
		"functionality is similar to Unix's ability to have multiple filing systems and discs to appear in any blank "
		"directory (eg; by using <tt>mount</tt>)<p>"
		"<u><b>WARNING:</b></u> This program uses functionality of NTFS undocumented publicly by Microsoft. While it is very unlikely "
		"data loss or misoperation on future versions of NTFS could occur, it still remains a remote possibility<p>"
		"<h2>To use:</h2>Very simply, set the destination directory where you want the symbolic links to be created. "
		"Then drag one or more directories or volumes (ie; entire drives) to the SymLink window, avoiding the red box. "
		"The symbolic links will be created, and the item marked <font color=\"#0000ff\">blue</font> so you know it's "
		"not a normal item (note it will only be blue if you have enabled \"Display compressed files and folders in an "
		"alternate colour\" which is available in Folder Options)<p>"
		"If you drag in a file, an alternative form of link is used which only works if the link is on the same drive "
		"as its source. Unlike Explorer shortcut files or directory links, file symbolic links are <b>identical</b> to "
		"the original file - in fact, there is one copy of the data with two identical links (the original and your new "
		"copy) pointing to the same data. Deleting a link does not destroy the data until all links to the data have been "
		"deleted. You can create a maximum of 1023 file links per NTFS volume (currently)<p>"
		"<b><u>WARNING:</u></b> Do NOT delete a symbolic link which points to a directory or volume using Explorer as normal "
		"as doing do will cause data loss (as I just found out by losing my BeOS installation <tt>:(</tt> ). What happpens "
		"is your data appears to be unmolested until you empty the Recycle Bin, at which stage all linked to data is "
		"destroyed. Hence to delete a link, just drag the link into the red box. Your linked to data will be delinked and "
		"hence be unaffected. Note deleting file links is deliberately unsupported due to potential for "
		"data loss - use Explorer as normal for this<p>"
		"v1.02 sets the system attribute for symbolic links to directories and volumes, so you will receive a warning from "
		"Explorer if you do try to delete it. This should prevent accidents.<p>"
		"<h2>Notes:</h2><ul>"
		"<li>The program is fully Unicode, so should work perfectly anywhere on the planet. Thanks Qt!"
		"<li>The links are stored as you specify them ie; using their x:\\ notation. Hence changing drive letters "
		"will cause the link to stop working (directory or volume links only)"
		"<li>Files cannot be linked across different volumes yet. It can be done, it just requires more grit (create "
		"a zero sized placeholder and reparse it to somewhere else). Feel free to add the functionality yourself if "
		"you really want it"
		"<li>Only items residing on the local machine can be linked. Network shares (whether by UNC notation or a "
		"share mapped to a drive letter) cannot be linked"
		"<li>There is a limitation of paths more than 260 characters. Since Explorer has the problem too we didn't "
		"worry about it too much"
		"</ul>"
		"<h2>History:</h2>"
		"<h3>v1.00 (11th July 2002):</h3>Released to the public<p>"
		"<h3>v1.01 (12th July 2002):</h3>Improved the help file. Added file links on the same drive using NTFS "
		"<tt>CreateHardLink()</tt>. Decreased executable size by 400Kb by optimising linker settings<p>"
		"<h3>v1.02 (13th July 2002):</h3>Emptied the Recycle Bin to get a strange error. Turned out all the symbolic links "
		"I had deleted using Explorer delete the linked data too (bye bye my BeOS installation). Issued immediate email "
		"warning to all testers and distribution groups and set to work on fixing it - after many hours experimenting, I "
		"found I can't stop Explorer wiping your files but I can have it show a warning about the link being a system file. My "
		"attempts to change permissions and security stopped the deletion but too late for the linked files! Also fixed "
		"bad tab order in widgets in main dialog.<p>"
		"<h2>Lastly:</h2>"
		"Thanks are due to Mikael Nordell (tamlin@algonet.se), some of his example link code was used verbatim (copyright "
		"remains with him) - you saved me a lot of work, thanks!<p>"
		"This program was designed and written using Qt made by Trolltech AS (http://www.trolltech.com/)<p>"
		"Niall Douglas<br>Madrid July 2002<p>SymLinkSupport@nedprod.com<br>http://www.nedprod.com/programs/Win32/SymLink "
		"</body></html>");
}

void SymLinkDialog::dragEnterEvent(QDragEnterEvent *e)
{
	e->accept(QUriDrag::canDecode(e));
}

void SymLinkDialog::dropEvent(QDropEvent *e)
{
	QStringList paths;
	if(QUriDrag::decodeLocalFiles(e, paths))
	{
		for(int n=0; n<paths.count(); n++)
		{	// Do the magic with paths[n] to DestPath->text()
			QString dest=DestPath->text();
			QChar src_c[MAX_PATH], dest_c[MAX_PATH];
			if(dest.endsWith("/") || dest.endsWith("\\")) dest.truncate(dest.length()-1);
			if(paths[n].length()==3 && paths[n][1]==':')
				dest+="\\"+paths[n].left(1);
			else
				dest+=paths[n].mid(paths[n].findRev('/'));
			Support_StringToCharArray(dest_c, QDir::convertSeparators(dest), MAX_PATH);
			Support_StringToCharArray(src_c, QDir::convertSeparators(paths[n]), MAX_PATH);
			if(GetFileAttributes((LPCTSTR) src_c) & FILE_ATTRIBUTE_DIRECTORY)
			{
				if(!CreateJunctionPoint((LPCTSTR) dest_c, (LPCTSTR) src_c)) goto decodeerror;
			}
			else
			{
				if(!CreateHardLink((LPCTSTR) dest_c, (LPCTSTR) src_c, NULL)) goto decodeerror;
			}
			if(0)
			{
				char errmsg[1024];
				int len;
				// Yeah, I know goto's are nasty ...
decodeerror:
				len=FormatMessageA(FORMAT_MESSAGE_FROM_SYSTEM | FORMAT_MESSAGE_IGNORE_INSERTS, 0, GetLastError(), 0, errmsg, 1024, 0);
				QMessageBox::critical(this, "Error", errmsg, QMessageBox::NoButton, QMessageBox::Ok, QMessageBox::NoButton);
			}
		}
	}
}

bool SymLinkDialog::eventFilter(QObject *o, QEvent *e)
{
	if(o==DestPath)
	{
		if(QEvent::Drop==e->type())
		{	// Need to override URI insert
			QDropEvent *de=(QDropEvent *) e;
			QStringList paths;
			if(QUriDrag::decodeLocalFiles(de, paths)) DestPath->setText(paths[0]);
			return true;
		}
	}
	else if(o==DeleteDrop)
	{
		if(QEvent::DragEnter==e->type())
		{
			QDragEnterEvent *de=(QDragEnterEvent *) e;
			de->accept(QUriDrag::canDecode(de));
			return true;
		}
		else if(QEvent::Drop==e->type())
		{
			QDropEvent *de=(QDropEvent *) e;
			QStringList paths;
			if(QUriDrag::decodeLocalFiles(de, paths))
			{
				QChar src_c[MAX_PATH];
				for(int n=0; n<paths.count(); n++)
				{
					Support_StringToCharArray(src_c, QDir::convertSeparators(paths[n]), MAX_PATH);
					if(!DeleteJunctionPoint((LPCTSTR) src_c))
					{
						char errmsg[1024];
						int len;
						len=FormatMessageA(FORMAT_MESSAGE_FROM_SYSTEM | FORMAT_MESSAGE_IGNORE_INSERTS, 0, GetLastError(), 0, errmsg, 1024, 0);
						QMessageBox::critical(this, "Error", errmsg, QMessageBox::NoButton, QMessageBox::Ok, QMessageBox::NoButton);
					}
				}
			}
			return true;
		}
	}
	return false;
}

void SymLinkDialog::openHomepage()
{
	SHELLEXECUTEINFO sei={ 0 };
	sei.cbSize=sizeof(sei);
	sei.fMask=0;
	sei.hwnd=NULL;
	sei.lpFile=L"http://www.nedprod.com/programs/Win32/SymLink/";
	ShellExecuteEx(&sei);
}

void SymLinkDialog::openHelp()
{
	help->show();
}

void SymLinkDialog::pathBrowse()
{
	DestPath->setText(QFileDialog::getExistingDirectory());
}
